--------------------------------------------------------------------------------
-- 
-- Name: WhoLibExports.lua
-- Author: Malex-Cenarius
-- Date: 12/23/2006
--
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
--
-- Method: WhoLib_GetUserInfo
--
-- Description: Retrieves user 'who' information either from the cache or
--	from the game.
--
-- Parameters:
--
--	userName - [in] The user name to query
--
--	maxTimestamp - [in, optional] Used to specify the tolerated age of the data
--		in seconds.  The default value is 1800 (30 min).  Any cached data that is
--		older than this value will be discarded and fresh data will be retrieved.
--		If you need the latest data, specify 0 here to clear the cached value, but 
--		don't call it repeatedly with 0, only once, otherwise no data will ever be
--		returned.
--
--  callback - [in, optional] An optional parameter that can be used to be notified 
--		of a completed query instead of polling.  If this parameter is specified
--		and WhoLib_GetUserInfo returns isValid=false, it means that the supplied
--		callback function will be called when the query is satisfied.
--
--		This function should be of the following form, where "data" is a table of 
--		user info values normally returned by WhoLib_GetUserInfo with the inclusion
--		of the "Name" value and the exlusion of the "timestamp" value.
--		
--		local function MyGetUserInfoCallback(data, context)
--
--			-- Use "context" to determine the pending request that the "data"
--			-- belongs to
--			--
--			
--			-- Access the user info through the "data" parameter
--			--
--			if(data ~= nil) then
--
--				Print(data.Name);
--				Print(data.GuildName);
--				Print(data.Level);
--				Print(data.Race);
--				Print(data.Class);
--				Print(data.Zone);
--
--			else
--
--				Print("User not found");
--
--			end
--
--		end
--
--		Note:  If WhoLib_GetUserInfo returns isValid=true, the callback will
--			not be called.
--
--	context - [in, optional] An optional parameter used in conjunction with the
--		callback parameter.  This value will be passed back to the caller through
--		the callback function.  If you have multiple outstanding requests, you 
--		might want to use the context parameter to store the userName, so you know
--		what data you are getting when the callback is called.  If this parameter 
--		isn't specified, the callback function will receive a nil value as it's
--		"context" parameter.
--
-- Returns:
--
--	isValid - true/false indicating whether the data is available for the userName.
--		If true, it means that the return values are valid.  False means that data
--		is still pending from the server and the method should be called until the
--		value is true.
--
--	guildName - The guild name of the character if a member or else an empty string
--
--	level - The level of the character
--
--	race - The race of the character
--
--	class - The class of the character
--
--	zone - The zone of the character at the time of the query
--
--	timestamp - The time when the cache entry was last updated.  This value is 
--		the number of seconds since Jan 1 1970.  Use "SecondsToTime" function to
--		convert this to hours/minutes friendly string, or do a time() - timestamp
--		to figure out how old the data is.
--
--------------------------------------------------------------------------------
function WhoLib_GetUserInfo(...)
	
	-- Extract parameters from vararg
	--
	local userName = select(1, ...);
	local userNameLower = strlower(userName);
	local maxTimestamp = WhoLib.Constants.MaxTimestamp;
	local outputToChatFrame = false;
	local insertAfterFirst = false;
	local maxRetries = WhoLib.Constants.MaxRetries;
	local callback = nil;
	local context = nil;
	
	-- Use default maxTimestamp if not specified
	--
	if(select("#", ...) >= 2) then			
		maxTimestamp = select(2, ...);				
	end	
	
	-- Extract the callback
	--
	if(select("#", ...) >=3) then
		callback = select(3, ...);
	end
	
	-- Extract the context
	--
	if(select("#", ...) >=4) then
		context = select(4, ...);
	end
	
	-- Internal only argument for outputing to chat frame on completion
	--
	if(select("#", ...) >= 5) then	
		outputToChatFrame = select(5, ...);	
	end
	
	if(select("#", ...) >= 6) then	
		insertAfterFirst = select(6, ...);	
	end	
	
	if(select("#", ...) >= 7) then	
		maxRetries = select(7, ...);	
	end		
		
	-- Check to see if there is already a request in the queue
	--
	if(WhoLib.Queue[userNameLower] ~= nil) then
	
		-- The request is pending, so just return false
		--
		WhoLib.Util.DebugOut("Request pending for: " .. userName);
		
		-- If it is a user request, make sure we output the results
		--
		if(outputToChatFrame) then		
			WhoLib.Queue[userName].OutputToChatFrame = outputToChatFrame;
		end
		
		if(insertAfterFirst) then	
			WhoLib.Util.MoveWhoRequestForward(userNameLower);
		end
		
		-- Make sure we are already tracking the callback
		--		
		WhoLib.Util.AddCallbackToQueue(userNameLower, callback, context);		
		
		return false;
	
	end	
	
	-- Flush the cached value for the userName if it exists and
	-- the timestamp on the cached item is greated than the specifed
	-- maxTimestamp
	--
	if(WhoLibCache[userNameLower] ~= nil and (time() - WhoLibCache[userNameLower].Timestamp) >= maxTimestamp) then
	
		WhoLib.Util.DebugOut("Flushing cached value for: " .. userName);
		WhoLibCache[userNameLower] = nil;
		WhoLibCache.Size = WhoLibCache.Size - 1;
	
	end

	-- Check to see if the userName is already in the cache
	--
	if(WhoLibCache[userNameLower] ~= nil) then
	
		local userData = WhoLibCache[userNameLower];
		
		-- Output to chat frame if necessary
		--
		if(outputToChatFrame) then
			WhoLib.Util.OutputWhoDataToChatFrame(userData.Name, userData.Level, userData.Race, userData.Class, userData.GuildName, userData.Zone);			
		end		
	
		-- Return the cached data for the user
		--
		return true, userData.GuildName, userData.Level, userData.Race, userData.Class, userData.Zone, userData.Timestamp;		
	
	end
	
	-- Add the username to the who queue
	--
	local timestamp = GetTime();	
	
	WhoLib.Queue[userNameLower] = {
		Timestamp = timestamp,
		IsComplex = false,
		Retries = maxRetries,
		OutputToChatFrame = outputToChatFrame,
		Callbacks = {}
	};
	WhoLib.QueueSize = WhoLib.QueueSize + 1;	

	-- Add the callback/context to the queued item
	--
	WhoLib.Util.AddCallbackToQueue(userNameLower, callback, context);			
	
	-- Add the user to the filterList for prioritization
	--
	tinsert(WhoLib.QueueFilterList, userNameLower);	
	
	-- If the request is a priority request, move it either
	-- to the front of the queue or if there is already one pending,
	-- to right after the first one
	--
	if(insertAfterFirst) then
		WhoLib.Util.MoveWhoRequestForward(userNameLower);
	end	

	WhoLib.Util.DebugOut("|cFFFFFF00Queueing request for: " .. userName .. " QueueSize: (" .. WhoLib.QueueSize .. ")");

	-- Service the queue immediately
	--
	WhoLib.Util.ProcessQueue(0);
	
	return false;
	
end

--------------------------------------------------------------------------------
--
-- Method: WhoLib_Query
--
-- Description: Used to query for who information using a filter
--
-- Parameters:
--
--	filter - [in] This filter string will be directly passed to the SendWho()
--		method and determines the results returned
--
--  callback - [in, optional] An optional parameter that can be used to be notified 
--		of a completed query instead of polling.  If this parameter is specified
--		and WhoLib_Query returns isValid=false, it means that the supplied
--		callback function will be called when the query is satisfied.
--
--		This function should be of the following form, where "data" is a table of 
--		tables containing query results matching the supplied query filter.  Each
--		entry contains the user info for a single user, similar to the return values
--		of WhoLib_GetUserInfo with the inclusion of the "Name" value and the exclusion
--		of the "timestamp" value.
--		
--		local function MyQueryCallback(data, context)
--
--			-- Use "context" to determine the pending request that the "data"
--			-- belongs to
--			--
--			
-- 			-- Loop over all the results
--			--
--			if(data ~= nil) then
--
--				for index, item in ipairs(data) do
--
--					Print(item.Name);
--					Print(item.GuildName);
--					Print(item.Level);
--					Print(item.Race);
--					Print(item.Class);
--					Print(item.Zone);
--
--				end
--
--			else
--
--				Print("No results found");
--
--			end
--
--		end
--
--		Note:  If WhoLib_Query returns isValid=true, the callback will not be 
--			called.
--
--	context - [in, optional] An optional parameter used in conjunction with the
--		callback parameter.  This value will be passed back to the caller through
--		the callback function.  If you have multiple outstanding requests, you 
--		might want to use the context parameter to store the userName, so you know
--		what data you are getting when the callback is called.  If this parameter 
--		isn't specified, the callback function will receive a nil value as it's
--		"context" parameter.
--
-- Returns:
--
--	isValid - true/false indicating whether the data is available for the query.
--		If true, it means that the return values are valid.  False means that data
--		is still pending from the server and the method should be called until the
--		value is true.
--
--  data - A list of data representing the results of the query.  Each value in
--		the list is identical to the data returned from WhoLib_GetUserInfo with the
--		inclusion of the "Name" field, and the exclusion of the "timestamp".
--
--		For example:
--
--			if(data ~= nil) then
--
--				Print("ResultsCount: " .. getn(data));
--
--				for index, item in ipairs(data) do
--
--						Print("Name: " .. item.Name);
--						Print(GuildName: " .. item.GuildName);
--						etc...			
--
--				end
--
--			else
--
--				Print("No results found");
--
--			end
--
-- Remarks:
--
--	Results from this function are not cached.  Once the function returns true,
--	a future call to the function will initiate another query.  Store the returned
--	data set in your addon and use it from there.
--
--------------------------------------------------------------------------------
function WhoLib_Query(...)
	
	local filter = select(1, ...);
	local filterLower = strlower(filter);
	local callback = nil;
	local context = nil;
		
	-- Extract the callback
	--
	if(select("#", ...) >=2) then
		callback = select(2, ...);
	end
	
	-- Extract the context
	--
	if(select("#", ...) >=3) then
		context = select(3, ...);
	end
	
	
	-- Check the query results table for a match
	--
	if(WhoLib.QueryResults[filterLower] ~= nil) then

		local results = WhoLib.QueryResults[filterLower];
	
		-- Remove the results set from the results table
		--
		WhoLib.QueryResults[filterLower] = nil;
		WhoLib.QueryResults.n = WhoLib.QueryResults.n - 1;
	
		WhoLib.Util.DebugOut("Results retrieved for query: " .. filter);
	
		-- Return the data set
		--
		return true, results.Data;
	
	end	
	
	-- Check to see if there is already a request in the queue
	--
	if(WhoLib.Queue[filterLower] ~= nil) then
		
		-- Make sure we are already tracking the callback
		--		
		WhoLib.Util.AddCallbackToQueue(filterLower, callback, context);		
	
		-- The request is pending, so just return false
		--
		WhoLib.Util.DebugOut("Request pending for: " .. filter .. " QueueSize: " .. WhoLib.QueueSize);
		return false;
	
	end	
		
	-- Add the filter to the who queue
	--
	WhoLib.Queue[filterLower] = {
		Timestamp = GetTime(),
		IsComplex = true,
		Retries = WhoLib.Constants.MaxRetries,
		OutputToChatFrame = false,
		Callbacks = {}
	};
	WhoLib.QueueSize = WhoLib.QueueSize + 1;	
	
	-- Add the callback/context to the queued item
	--
	WhoLib.Util.AddCallbackToQueue(filterLower, callback, context);				
	
	-- Add the filter to the queue list
	--
	tinsert(WhoLib.QueueFilterList, filterLower);

	WhoLib.Util.DebugOut("|cFFFFFF00Queueing complex request for: " .. filter .. " QueueSize: (" .. WhoLib.QueueSize .. ")");

	-- Service the queue immediately
	--
	WhoLib.Util.ProcessQueue(0);
	
	return false;



end